(* Part 1: 
- introduction to inductive types and lists
- examples of recursive functions using pattern matching
- an inductive proof example *)

(* Bonus content:
- a proof of the pigeonhole principle*)

Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Arith.EqNat.
Require Import Coq.omega.Omega.

Module ListFundamentals.

Inductive list (X:Type) : Type :=
  | nil : list X
  | cons : X -> list X -> list X.

Compute(cons nat 3 (nil nat)).

Arguments nil {X}.
Arguments cons {X} _ _.

Compute(@nil nat).
Compute(cons 3 nil).

Fixpoint app {X : Type} (l1 l2 : list X)
             : (list X) :=
  match l1 with
  | nil      => l2
  | cons h t => cons h (app t l2)
  end.

Notation "x :: y" := (cons x y)
                     (at level 60, right associativity).
Notation "[ ]" := nil.
Notation "[ x ; .. ; y ]" := (cons x .. (cons y []) ..).
Notation "x ++ y" := (app x y)
                     (at level 60, right associativity).


Fixpoint Inb (x : nat) (l : list nat) : bool :=
  match l with
  | [] => false
  | x' :: l' => (beq_nat x x') || Inb x l'
  end.

Fixpoint In {A : Type} (x : A) (l : list A) : Prop :=
  match l with
  | [] => False
  | x' :: l' => x' = x \/ In x l'
  end.

End ListFundamentals.

Require Import Coq.Lists.List.
Import ListNotations.

(** First prove an easy useful lemma. *)

Lemma in_split : forall (X:Type) (x:X) (l:list X),
  In x l ->
  exists l1 l2, l = l1 ++ x :: l2.
Proof. {
    intros X x. induction l as [ | a l IHl].
    { intros H1. inversion H1. }
    { simpl. intros [H2 | H3].
      { exists []. exists l. simpl. rewrite -> H2. reflexivity. }
      { apply IHl in H3. destruct H3 as [l1' [l2' H4]].
        exists (a::l1'). exists l2'. rewrite -> H4. reflexivity. }
    }
  } Qed.

(** Now define a property [repeats] such that [repeats X l] asserts
    that [l] contains at least one repeated element (of type [X]).  *)

Inductive repeats {X:Type} : list X -> Prop :=
| repeats_case: forall (l1 l2 l3:list X) (x: X), repeats (l1 ++ x :: l2 ++ x :: l3).

Definition pigeonhole_principle_statement: Prop :=
  forall (X:Type) (l1 l2:list X),
    (forall x, In x l1 -> In x l2) ->
    length l2 < length l1 ->
    repeats l1.

Lemma pigeonhole_principle_lemma_1: forall {X: Type} (l1 l2 u: list X),
                                      (forall x, In x l1 -> In x (u ++ l2)) ->
                                      (length l2 < length l1) -> repeats (u ++ l1).
Proof. {
    intros X. induction l1 as [| x l1 IHl1].
    { (* l1 = [] *) simpl. intros l2 u H1 H2. inversion H2. }
    { (* l1 = x::l1' *)
      intros l2 u H3 H4. assert (In x u \/ In x l2) as [H5 | H6].
      { (* proving: In x u \/ In x l2 *)
        Check in_app_iff. apply in_app_iff. apply H3. simpl. left. reflexivity. }
      { (* In x u *)
        assert (exists l0 l3 : list X, u = l0 ++ x :: l3) as [l0 [l3 H7]].
        { (* proving: exists l0 l3 : list X, u = l0 ++ x :: l3 *)
          Check in_split. apply in_split. apply H5. }
        Check @repeats_case.
        rewrite -> H7. replace ((l0 ++ x :: l3) ++ x :: l1) with (l0 ++ x :: l3 ++ x :: l1).
        apply (repeats_case l0 l3 l1 x).
        { (* proving: l0 ++ x :: l3 ++ x :: l1 = (l0 ++ x :: l3) ++ x :: l1 *)
          rewrite <- app_assoc. reflexivity. }
      }
      { (* In x l2 *)
        assert (exists l0 l3 : list X, l2 = l0 ++ x :: l3) as [l0 [l3 H8]].
        { (* proving: exists l0 l3 : list X, l2 = l0 ++ x :: l3 *)
          apply in_split, H6. }
        replace (u ++ x :: l1) with ((u ++ [x]) ++ l1).
        apply (IHl1 (l0 ++ l3) (u ++ [x])).
        { (* proving: forall x0 : X, In x0 l1 -> In x0 ((u ++ [x]) ++ l0 ++ l3) *)
          intros x0 H9. apply in_app_iff. assert (In x0 u \/ In x0 l2) as [H10 | H11].
          { (* proving: In x0 u \/ In x0 l2 *)
            apply in_app_iff, H3. simpl. right. apply H9. }
          { (* In x0 u *)
            left. apply in_app_iff. left. apply H10. }
          { (* In x0 l2 *)
            assert (In x0 l0 \/ In x0 (x :: l3)) as [H12 | H13].
            { (* proving: In x0 l0 \/ In x0 (x :: l3) *)
              apply in_app_iff. rewrite <- H8. apply H11. }
            { (* In x0 l0 *)
              right. apply in_app_iff. left. apply H12. }
            { (* In x0 (x :: l3) *)
              simpl in H13. destruct H13 as [H14 | H15].
              { (* x = x0 *)
                left. apply in_app_iff. right. simpl. left. apply H14. }
              { (* In x0 l3 *)
                right. apply in_app_iff. right. apply H15. }
            }
          }
        }
        { assert (length l2 = S (length l0 + length l3)) as H9.
          { (* proving: length l2 = S (length l0 + length l3) *)
              rewrite -> H8, -> app_length. simpl. symmetry. Check plus_n_Sm. apply plus_n_Sm. }
          unfold lt. apply le_S_n. rewrite -> app_length, <- H9. simpl in H4.  apply H4. 
        }
        { (* proving: (u ++ [x]) ++ l1 = u ++ x :: l1 *)
          rewrite <- app_assoc. reflexivity. }
      }
    }
  } Qed.

Theorem pigeonhole_principle: pigeonhole_principle_statement.
Proof. {
    intros X l1 l2 H1 H2. apply (pigeonhole_principle_lemma_1 l1 l2 []).
    { apply H1. }
    { apply H2. }
  } Qed.

(** $Date: 2015-08-11 12:03:04 -0400 (Tue, 11 Aug 2015) $ *)
